def remove_prefix(x: str, y: str) -> str:
    """ to be replaced with str.removeprefix when python >= 3.9 """
    if x.startswith(y):
        x = x[len(y):]
    return x


def remove_suffix(x: str, y: str) -> str:
    """ to be replaced with str.removesuffix when python >= 3.9 """
    assert len(y) > 0
    if x.endswith(y):
        x = x[:-len(y)]
    return x



from functools import lru_cache
from pathlib import Path
from typing import List, Dict, Tuple, Any

import toml

from newpyter.storage import AbstractStorage, Encryptor, SSHStorage, S3Storage, LocalStorage, MemoryMockStorage


class NewpyterConfigError(RuntimeError):
    pass


def _get_storage_parameters(notebook_filename: Path, storage_url: str = None) -> Tuple[str, str, Dict[str, Any]]:
    """
    Parses settings and finds any additional configuration parameters for the storage
    If storage url is known, will return parameters for it, otherwise will
    """
    nb_path = Path(notebook_filename)
    local_cache_dir = None
    # local is always available
    storages = {'local': {}}

    found_config_files = []
    possible_config_paths = [p.joinpath('.newpyter_config.toml') for p in [*nb_path.parents, Path.home()]]
    for config_path in possible_config_paths:
        if config_path.exists():
            found_config_files.append(config_path)
            with config_path.open() as f:
                try:
                    config = toml.load(f).get('newpyter', {})
                except BaseException as e:
                    raise NewpyterConfigError(f'Error loading config {config_path}: {e}')

                if storage_url is None:
                    storage_url = config.get('default_storage', None)
                if local_cache_dir is None:
                    local_cache_dir = config.get('local_cache_dir', None)
                # already known storages have precedence
                storages = {**config.get('storages', {}), **storages}

        if storage_url is not None and storage_url in storages and local_cache_dir is not None:
            # no need to traverse further, we already ide
            storage_settings = storages[storage_url]
            return storage_url, local_cache_dir, storage_settings

    if storage_url is not None and local_cache_dir is not None:
        # storage is set, but no additional parameters were found
        return storage_url, local_cache_dir, {}

    def format_paths_absolute(paths: List[Path]) -> str:
        result = ''
        for p in paths:
            result += f'  {p.absolute()}\n'
        return result

    message = f"""
Newpyter could not locate any storage. Following shouldn't be None: 
default_storage = {storage_url};
local_cache_dir = {local_cache_dir};

Make sure you have properly setup .newpyter_config.toml

Scanned following locations:
{format_paths_absolute(possible_config_paths)} 

Found following locations:
{format_paths_absolute(found_config_files)}

Parsed settings: 

default storage: {storage_url}
found storages: 
{list(storages)}
"""

    raise NewpyterConfigError(message)


class HashableDict:
    def __init__(self, d: Dict):
        self.internals = frozenset(d.items())

    def __hash__(self):
        return hash(self.internals)

    def to_dict(self):
        return dict(list(self.internals))


@lru_cache(typed=True)
def _get_storage(url: str, local_cache_dir: str, other_params: HashableDict) -> AbstractStorage:
    """
    Storage for parameters, intentionally cached to keep the same storage object
    if parameters were not updated
    """
    other_params = other_params.to_dict()
    password = other_params.pop('password', None)
    encryptor = Encryptor(password=password)
    local_cache_dir = Path(local_cache_dir).expanduser()
    if url.startswith('ssh://'):
        return SSHStorage(url, encryptor=encryptor, local_cache=local_cache_dir, **other_params)
    if url.startswith('s3://'):
        return S3Storage(url, encryptor=encryptor, local_cache=local_cache_dir, **other_params)
    if url == 'local':
        return LocalStorage(encryptor=encryptor, local_cache=local_cache_dir, **other_params)
    if url == ':MemoryMockStorage:':
        return MemoryMockStorage()
    raise NewpyterConfigError(f"Did not recognize URL for storage manager: {url}")


def get_storage_for_notebook(notebook_filename: Path, storage_url: str = None) -> AbstractStorage:
    """
    :param notebook_filename:
        path to notebook on FS.
        It is important as config files are search for in parent folders
    :param storage_url:
        if storage URL is known, configs are parsed for additional parameters.
        otherwise, URL is found by looking for default storage and corresponding parameters
    """
    url, local_cache_dir, other_params = _get_storage_parameters(notebook_filename, storage_url=storage_url)
    return _get_storage(url, local_cache_dir=local_cache_dir, other_params=HashableDict(other_params))



from pathlib import Path

import nbformat
from jupyter_server.services.contents.largefilemanager import LargeFileManager
from tornado.web import HTTPError

from .config import get_storage_for_notebook
from .parsing import decompose_notebook, recompose_notebook


class NewpyterContentsManager(LargeFileManager):
    """
    Contents manager that can handle `newpyter` notebooks.
    Aside from representing newpyter notebooks as text (and back),
    they are responsible for conversion and listing/editing/renamings of files
    """

    def __init__(self, **kargs):
        super().__init__(**kargs)
        self.log.info(f'Newpyter activated')

    def get(self, path, content=True, type=None, format=None):
        if path.endswith('.newpy') and type in (None, 'notebook'):
            try:
                return self._newpy_notebook_model(path, content=content)
            except BaseException as e:
                raise HTTPError(500, f'Could not load notebook: \n{e}')
        else:
            return LargeFileManager.get(self, path=path, content=content, type=type, format=format)

    def _newpy_notebook_model(self, path, content=True):
        """Build a notebook model

        if content is requested, the notebook content will be populated
        as a JSON structure (not double-serialized)
        """
        model = self._base_model(path)
        model['type'] = 'notebook'
        os_path = self._get_os_path(path)

        if content:
            with open(os_path, encoding='utf-8') as f:
                nb = recompose_notebook(f.read(), notebook_path=os_path)

            # just mark all cells as trusted
            self.notary.mark_cells(nb, trusted=True)
            model['content'] = nb
            model['format'] = 'json'
            self.validate_notebook_model(model)
        return model

    def save(self, model, path=""):
        """Save the file model and return the model with no content."""
        if not (path.endswith('.newpy') and model['type'] == 'notebook'):
            return LargeFileManager.save(self, model=model, path=path)

        try:
            os_path = self._get_os_path(path)

            self.log.debug("Saving newpyter notebook %s", os_path)

            decomposed = decompose_notebook(
                notebook=nbformat.from_dict(model['content']),
                storage=get_storage_for_notebook(notebook_filename=Path(os_path)),
            )

            with self.atomic_writing(os_path, encoding='utf-8') as f:
                f.write(decomposed)

            model = self.get(path, content=False)
            # ignore post hooks for newpyter notebooks
            # self.run_post_save_hook(model=model, os_path=os_path)
            return model

        except Exception as e:
            self.log.error(u"Error while saving file: %s %s", path, e, exc_info=True)
            raise HTTPError(500, u"Unexpected error while saving file: %s %s" % (path, e))

    def rename_file(self, old_path, new_path) -> None:
        self.log.error(f'Move called: {old_path} {new_path}')

        if Path(old_path).suffix == '.newpy' and Path(new_path).suffix == '.ipynb':
            return self.convert_from_newpy_to_ipynb_or_reverse(old_path, new_path)

        if Path(new_path).suffix == '.newpy' and Path(old_path).suffix == '.ipynb':
            return self.convert_from_newpy_to_ipynb_or_reverse(old_path, new_path)

        return super().rename_file(old_path, new_path)

    def convert_from_newpy_to_ipynb_or_reverse(self, old_path, new_path) -> None:
        self.log.info(f'Conversion from {old_path} to {new_path}')
        if Path(self._get_os_path(new_path)).exists():
            raise HTTPError(409, u'File already exists: %s' % new_path)
        if not Path(self._get_os_path(old_path)).exists():
            raise HTTPError(404, u'File does not exist: %s' % old_path)

        model = self.get(old_path)
        self.save(model, path=new_path)
        self.delete(old_path)



"""
Ugly and terribly simple viewer for newpyter notebook.
It can show ipynb too, as well as traverse the github (given that API token is provided).

See /demo/viewer for instructions how to start this server

Main TODOs
- better app styling
- reduce code ugliness by using templates

"""
import html
import tempfile
from pathlib import Path
from typing import Tuple, Optional

import sh
from fastapi import FastAPI
from fastapi.responses import HTMLResponse
from nbconvert import HTMLExporter

from newpyter.ContentsManager import NewpyterContentsManager

style = """
<style>
body {
    min-width: 800px;
    max-width: 1200px;
    margin: 50px auto;
}
.collapsed-input {
    max-height: 32px;
    opacity: 0.5;
}
.jp-RenderedImage img.full-width-image, img.full-width-image {
    max-width: initial;
}
</style>
"""


class NewpyRouter:
    def __init__(self, github_token: str, start_repo: str, newpyter_toml_path: Path, web_basedir='/'):
        self.github_token = github_token
        self.start_repo = start_repo
        self.newpyter_toml_path = newpyter_toml_path
        self.web_basedir = web_basedir

        self.app = FastAPI()
        self.router = self.app
        self.subapp = FastAPI()
        self.app.mount(web_basedir, self.subapp)

        @self.subapp.get('/', response_class=HTMLResponse)
        def frontpage() -> str:
            return self.frontpage()

        @self.subapp.get('/source', response_class=HTMLResponse)
        def read_source(source: str) -> str:
            return self.read_source(source=source)

        @self.subapp.get('/github', response_class=HTMLResponse)
        def read_github(github_permalink: str) -> str:
            return self.read_github(github_permalink=github_permalink)

        @self.subapp.get('/github_dir', response_class=HTMLResponse)
        def list_github(repo: str = '', path_in_repo: str = ''):
            return self.list_github(repo=repo, path_in_repo=path_in_repo)

    def frontpage(self):
        return style + render_form(app=self.router, default_repo=self.start_repo)

    def read_source(self, source: str) -> str:
        return render_form(source=source, app=self.router) + convert_newpy_text_to_html(source) + style

    def read_github(self, github_permalink: str) -> str:
        if "github" in github_permalink:
            html, newpyter_source = download_and_convert(github_permalink, secret=self.github_token,
                                                         newpyter_toml=self.newpyter_toml_path)
            contents = style + render_form(url=github_permalink, source=newpyter_source, app=self.router,
                                           collapse_code=True) + html
            return contents
        else:
            return f"Did not recognize link: {render_form(url=github_permalink, app=self.router) + github_permalink}"

    def list_github(self, repo: str = '', path_in_repo: str = '') -> str:
        result = download_and_render_link_list(
            github_token=self.github_token,
            repo=self.start_repo if repo == '' else repo,
            path_in_repo=path_in_repo,
            app=self.router,
        ) or ''
        return f'<pre>{result}</pre>'


def render_form(url="", source="", default_repo='', collapse_code=False, *, app) -> str:
    """Allows selecting GitHub link or write source"""
    collapse_source = "collapseSource();" if collapse_code else ""

    js_content = """
    <script>
    var collapsible_code_elements = document.getElementsByClassName("jp-Editor");

    function onload() {
        collapsible_code_elements = document.getElementsByClassName("jp-Editor");
        for(let elem of collapsible_code_elements) {
            let e = elem;
            e.addEventListener('dblclick', function(event){e.classList.toggle('collapsed-input');});
        };
        var output_images = document.querySelectorAll(".jp-RenderedImage img");
        for(let elem of output_images) {
            let e = elem;
            e.addEventListener('dblclick', function(event){console.log(e); e.classList.toggle('full-width-image');});
        };
        // collapse_code_insertion
    };
    window.addEventListener('load', onload);
    function collapseSource() {
        for(let elem of collapsible_code_elements) {
            console.log('collapsing', elem);
            elem.classList.add('collapsed-input');
        }
    };
    function showSource() {
        for(let elem of collapsible_code_elements) {
            console.log('un-collapsing', elem);
            elem.classList.remove('collapsed-input');
        }
    };
    </script>
    """.replace('// collapse_code_insertion', collapse_source)

    return f"""
    <div style='margin-bottom: 15px;'>
        <label for='nothing_just_spacer' style='width: 170px; display: inline-block;'></label>
        {make_button_to_get_current_link()}
        &nbsp;
        <button type="button" onclick="collapseSource()">minimize code</button>
        &nbsp;
        <button type="button" onclick="showSource()">maximize code</button>
        &nbsp;
        Double click on input to minimize/maximize individual code input
    </div>
    <form action="{app.url_path_for('read_github')}" method='get'>
      <label for='github_permalink' style='width: 170px; display: inline-block;'>Github permalink</label>
      <input type="text" id="github_permalink" name="github_permalink" value="{html.escape(url)}" size=100 >
      <input type="submit" value="Submit">
    </form>
    <form action="{app.url_path_for('read_source')}" method='get'>
      <label for='github_permalink' style='width: 170px; display: inline-block;' >Or source code:</label>
      <textarea id="source" name="source" rows="4" cols="100"  style='vertical-align: top;'>{html.escape(source)}</textarea>
      <input type="submit" value="Submit">
    </form>
    <form action="{app.url_path_for('list_github')}" method='get'>
      <label for='repo' style='width: 170px; display: inline-block;' >Or repository:</label>
      <input type="text" id="repo" name="repo" value="{default_repo}" size=100 >
      <input type="submit" value="Submit">
    </form>    
    
    {js_content}
    """


def make_link(app, method, **kargs) -> str:
    from urllib.parse import urlencode
    return app.url_path_for(method) + '?' + urlencode(kargs)


def convert_newpy_text_to_html(source: str):
    with tempfile.TemporaryDirectory() as f:
        # copy config
        # with Path(f).joinpath(".newpyter_config.toml").open("w", encoding="utf-8") as ftarget:
        #     ftarget.write(newpyter_config_contents)

        local_path = Path(f).joinpath("nb.newpy")
        with local_path.open("w") as nbf:
            nbf.write(source)

        manager = NewpyterContentsManager()
        manager.root_dir = "/"
        old_local_path = local_path
        local_path = old_local_path.with_suffix(".ipynb")
        manager.convert_from_newpy_to_ipynb_or_reverse(str(old_local_path.absolute()), str(local_path.absolute()))

        html_exporter = HTMLExporter()
        html_exporter.template_name = "classic"

        # 3. Process the notebook we loaded earlier
        (body, resources) = html_exporter.from_filename(
            str(local_path),
        )
        return body


def make_button_to_get_current_link() -> str:
    return """
    <button onclick="copyToClip(document.getElementById('div-with-link').innerHTML)">Copy link to this page</button>
    <div id="div-with-link" style="display:none">
        <a id='link_to_viewer' href=''>newpy viewer</a>
    </div>
    <script>
        document.getElementById('link_to_viewer').href = window.location.href;
        function copyToClip(str) {
          function listener(e) {
            e.clipboardData.setData("text/html", str);
            e.clipboardData.setData("text/plain", str);
            e.preventDefault();
          }
          document.addEventListener("copy", listener);
          document.execCommand("copy");
          document.removeEventListener("copy", listener);
        };
    </script>
    """


def download_and_convert(permalink: str, secret: str, newpyter_toml: Path) -> Tuple[str, str]:
    assert permalink.startswith("https://github.com/")
    organization, repository, _blob_, commit, *path = permalink[len("https://github.com/"):].split("/")

    with tempfile.TemporaryDirectory() as f:
        # copy config in case conversion is required
        with Path(f).joinpath(".newpyter_config.toml").open("w", encoding="utf-8") as ftarget:
            ftarget.write(newpyter_toml.read_text())

        local_path = Path(f).joinpath(path[-1])

        path_in_repo = "/".join(path)
        request_link = f"https://api.github.com/repos/{organization}/{repository}/contents/{path_in_repo}?ref={commit}"

        print(request_link)

        sh.curl(
            "-H",
            "Accept: application/vnd.github.v4.raw",
            "-H",
            f"Authorization: Bearer {secret}",
            L=request_link,
            o=local_path,
        )
        contents = open(local_path).read()

        print(contents)

        if '"message": "Not Found",' in contents:
            print(contents)
            return contents, ""

        if local_path.name.endswith(".newpy"):
            manager = NewpyterContentsManager()
            manager.root_dir = "/"
            # for some reason on mac m1 is does not work without allow_hidden
            manager.allow_hidden = True
            old_local_path = local_path
            local_path = old_local_path.with_suffix(".ipynb")
            # assert 0 == 1, old_local_path
            assert old_local_path.exists()
            manager.convert_from_newpy_to_ipynb_or_reverse(str(old_local_path.absolute()), str(local_path.absolute()))

        if local_path.name.endswith(".ipynb"):
            # 1. Instantiate the exporter. We use the `classic` template for now; we'll get into more details
            # later about how to customize the exporter further.
            html_exporter = HTMLExporter()
            html_exporter.template_name = "classic"

            # 2. Process the notebook we loaded earlier
            (body, resources) = html_exporter.from_filename(
                str(local_path),
            )
            return body, contents

        if local_path.name.endswith(".html"):
            return contents, ""

        raise NotImplementedError()


def download_and_render_link_list(github_token, repo, path_in_repo, *, app) -> Optional[str]:
    with tempfile.TemporaryDirectory() as d:
        local_path = Path(d).joinpath('contents')
        sh.curl(
            "-H",
            "Accept: application/vnd.github.v4.raw",
            "-H",
            f"Authorization: Bearer {github_token}",
            L=f'https://api.github.com/repos/{repo}/contents/{path_in_repo}',
            o=local_path,
        )
        file_contents = open(local_path).read()
        try:
            import json
            items_in_folder: list[dict] = json.loads(file_contents)
            items_in_folder = list(items_in_folder)
        except:
            return None

        assert len(items_in_folder) > 0

        result = ''
        if path_in_repo.endswith('/'):
            path_in_repo = path_in_repo[:-1]

        if path_in_repo != '':
            parent_path = str(Path(path_in_repo).parent)
            link = make_link(app, 'list_github', repo=repo, path_in_repo=parent_path)
            result += f'<a href="{link}">..</a> <br/>'

        for item in items_in_folder:
            name = item['name']
            if item['type'] == 'dir':
                link = make_link(app, 'list_github', repo=repo, path_in_repo=path_in_repo + '/' + name)
                result += f'<a href="{link}">{name}</a> <br/>'
            elif item['type'] == 'file':
                if name.endswith('.ipynb') or name.endswith('.newpy'):
                    link = make_link(app, 'read_github', github_permalink=item['html_url'])
                    result += f'<a href="{link}">{name}</a> <br/>'
                else:
                    # do not provide links for other files
                    result += f'{name} <br/>'
            else:
                result += f'{name} (unrecognized) <br/>'

            # import pprint
            # result += pprint.pformat(item) + '<br />'

        result = f'''
        <div style='margin: 200px 200px;'>
        <h2>{path_in_repo}</h2>
        <p style='font-size: 14px;'>{result}</p>
        </div>
        '''
        return result





usage = """
Usages:
python -m newpyter --to ipynb file1.newpy file2.newpy
python -m newpyter --to newpy file1.ipynb file2.ipynb
python -m newpyter debug   # shows used storage in cwd 

(if your env with newpyter is activated, you can skip 'python -m')
"""
# CLI for newpyter
import sys
from pathlib import Path

from newpyter.ContentsManager import NewpyterContentsManager


def to_paths(files, suffix):
    files = [Path(f).absolute() for f in files]
    for f in files:
        assert f.exists(), f'file {f} does not exist'
        assert f.suffix == suffix, (f.suffix, suffix)
    return files


manager = NewpyterContentsManager()
manager.root_dir = '/'



def main():
    if sys.argv[1:] == ['debug']:
        from newpyter.config import get_storage_for_notebook
        storage = get_storage_for_notebook(notebook_filename=Path().joinpath('nonexistent.ipynb'))
        print(storage)
        return

    if len(sys.argv) < 4:
        print(usage)
        exit(1)

    _, command, output_format, *files = sys.argv
    if command != '--to' or (output_format not in ['ipynb', 'newpy']):
        print(usage)
        exit(1)

    input_suffix = {'ipynb': '.newpy', 'newpy': '.ipynb'}[output_format]

    files = to_paths(files, input_suffix)
    for file in files:
        print('processing', file)
        manager.convert_from_newpy_to_ipynb_or_reverse(
            str(file),
            str(file.with_suffix(f'.{output_format}')),
        )


if __name__ == '__main__':
    main()



from pathlib import Path
__version__ = '0.5.2'

def patch_jupyterlab_frontend(allow_patching_multiple=True):
    """
    Jupyter's frontend code does not recognize any other extension but ipynb as notebooks.
    No matter what jupyter backend informs and passes.

    In jlab 4.X there seems to be a way to modify this behavior with labextension,
    but I don't want to maintain full extension given we basically need to replace one line js.
    """
    try:
        from jupyterlab._version import __version__
    except ImportError:
        print('Did not locate jupyterlab')
        # no jupyter lab - no problems. Jupyter notebook seems to work properly
        return
    from jupyter_core.paths import jupyter_path

    candidate_files = [
        Path(file)
        for path in jupyter_path()
        for file in Path(path).glob('**/*.js')
    ]

    def contains_one_of(source_text, options):
        return any(option in source_text for option in options)

    original = '[".ipynb"]'
    patched = '[".ipynb", ".newpy"]'

    appropriate_files = [
        p for p in candidate_files
        if p.parent.name == 'static'
        and contains_one_of(p.open().read(), [original, patched])
    ]
    if len(appropriate_files) == 0:
        raise RuntimeError(f'Found no appropriate js to patch among {candidate_files} in {jupyter_path()}')
    if len(appropriate_files) > 1 and not allow_patching_multiple:
        raise RuntimeError(f'Found several appropriate js files to patch: {appropriate_files}')

    for file in appropriate_files:
        with file.open() as f:
            text = f.read()

        if original in text:
            print(f"Patching jupyterlab to support .newpy extension at {file}")
            with file.open('w') as f:
                f.write(text.replace(original, patched))
            print(f"Patched  jupyterlab to support .newpy extension at {file}")
        elif patched in text:
            print(f"Jupyterlab is already patched at {file}")
        else:
            raise RuntimeError('Did not find fragment to be patched in jupyterlab')



"""
This module parses and saves a notebook.
These two processes are kept together for a simpler testing and better parallel.

Parsing uses grammar parsing provided by parsimonous package.
You may note grammar_text variable filled with parts of grammar along with code that reconstructs a notebook.

"""

import ast
import re
from pathlib import Path
from platform import python_version
from typing import List, Tuple, Dict

import nbformat
from nbformat import NotebookNode
from parsimonious import ParseError
from parsimonious.grammar import Grammar, NodeVisitor
from tornado.web import HTTPError

from .config import get_storage_for_notebook
from .storage import AbstractStorage, StorageContext
from .utils import remove_suffix


def encode_cell_text(string: str, cell_type='md'):
    assert cell_type in ['md', 'raw']
    encoded = repr(string)[1:-1]  # remove single quotes, this escapes string
    encoded = encoded.replace(r"\'", "'").replace('"', r'\"')
    # replacing \n that is not preceded with one more \ with another real thing
    # should be rewritten with a simple function
    encoded = re.sub(r'(?<!\\)\\n', '\n', encoded)
    assert '"""' not in encoded
    evaluated = ast.literal_eval(f'"""{encoded}\n"""')
    assert evaluated == string + '\n', (string, evaluated, encoded)
    return f'"""{cell_type}\n{encoded}\n"""'


def decode_cell_text(string):
    # parse literal using python built-in method
    string_with_additional_lines = ast.literal_eval(string)
    assert isinstance(string_with_additional_lines, str)
    cell_type, *content = remove_suffix(string_with_additional_lines, '\n').split('\n')
    cell_type = str.strip(cell_type)
    return dict(
        cell_type={'md': 'markdown', 'raw': 'raw'}[cell_type],
        metadata={},
        source='\n'.join(content),
    )


class NewpyterOutputPlaceholder:
    def __init__(self, hashes):
        self.hashes = hashes


class NotebookVisitor(NodeVisitor):
    newline = '\n'

    # grammar is composed line-by-line, each line is kept next to a visitor
    grammar_text = ""
    grammar_text += r'notebook = config cells ~r"(\s*\n)*" ' + newline

    def visit_notebook(self, node, visited_children):
        config, cells, _useless_tail = visited_children
        return {
                   'cells': cells,
                   # inserting metadata for a notebook.
                   # It may be not coinciding with what jupyter considers right
                   # in this case it will constantly report that the notebook is changed
                   # need a better way to handle this in the future
                   'metadata': {'kernelspec': {'display_name': 'Python 3',
                                               'language': 'python',
                                               'name': 'python3'},
                                'language_info': {'codemirror_mode': {'name': 'ipython',
                                                                      'version': 3},
                                                  'file_extension': '.py',
                                                  'mimetype': 'text/x-python',
                                                  'name': 'python',
                                                  'nbconvert_exporter': 'python',
                                                  'pygments_lexer': 'ipython3',
                                                  'version': python_version()}},
                   'nbformat': 4,
                   'nbformat_minor': 0,
               }, config

    grammar_text += r'config = config_line* ' + newline

    def visit_config(self, node, visited_children):
        return [(k, v) for k, v in visited_children]

    grammar_text += r'config_line = "# newpyter." ~r"[a-zA-Z\.]+" ~r" ?: ?" ~r"[^\n]*\n" ' + newline

    def visit_config_line(self, node, visited_children):
        _, key, _, value = visited_children
        return key.text.strip(), value.text.strip()

    grammar_text += r'cells = cell* ' + newline

    def visit_cells(self, node, visited_children):
        return list(visited_children)

    grammar_text += r'cell = comment_like_cell/code_cell' + newline

    def visit_cell(self, node, visited_children):
        return visited_children[0]

    grammar_text += r'code_cell = code_cell_header (out_line/code_line)*' + newline

    def visit_code_cell(self, node, visited_children):
        execution_count, parsed_lines = visited_children
        code = []
        out_hashes = []
        for [[type, content]] in parsed_lines:
            if type == 'code':
                code.append(content)
            else:
                assert type == 'out'
                if content is None:
                    continue
                out_hashes.append(content)

        # last empty line added in newpy just for readability
        code = remove_suffix('\n'.join(code), '\n')

        result = dict(
            cell_type='code',
            outputs=NewpyterOutputPlaceholder(hashes=out_hashes),
            # make trusted by default
            metadata={},
            source=code,
            execution_count=execution_count,
        )

        return result

    grammar_text += r'code_cell_header = ~r"# In\[" exec_code ~r" ?\] ?\n" ' + newline

    def visit_code_cell_header(self, node, visited_children):
        _, exec_code, _ = visited_children
        return exec_code

    grammar_text += r'exec_code = ~r"\d*" ' + newline

    def visit_exec_code(self, node, visited_children):
        content = node.text
        return None if content == "" else int(content)

    grammar_text += r'code_line = !code_cell_header !comment_cell_header !out_line ~r"[^\n]*\n" ' + newline

    def visit_code_line(self, node, visited_children):
        assert isinstance(node.text, str)
        return 'code', node.text.rstrip('\n')

    grammar_text += r'out_line  = ~r"# Out: ?" output_code ~r" ?\n" ' + newline

    def visit_out_line(self, node, visited_children):
        _, output_code, _ = visited_children
        return 'out', output_code

    grammar_text += r'output_code = ("empty" / ~r"[0-9a-f]+") ' + newline

    def visit_output_code(self, node, visited_children):
        content = node.text
        return None if content == 'empty' else content

    grammar_text += r'''
    comment_like_cell = comment_cell_header (!comment_cell_tail ~r"[^\n]*\n")* comment_cell_tail
    comment_cell_header = ~r'""" ?' ("md" / "raw") "\n"
    comment_cell_tail = ~r'""" ?\n+'
    '''.strip("\n ") + newline

    def visit_comment_like_cell(self, node, visited_children):
        return decode_cell_text(node.text)

    def generic_visit(self, node, visited_children):
        """ The generic visit method, applied to everything else """
        return visited_children or node


notebook_grammar = Grammar(NotebookVisitor.grammar_text)


def check_recursively_json(node, transformation):
    """ Traverse parsed json for notebook and apply transformation to each node """
    assert isinstance(node, (int, float, list, dict, str, type(None), NewpyterOutputPlaceholder)), (node, type(node))
    if isinstance(node, list):
        node = [check_recursively_json(x, transformation) for x in node]

    if isinstance(node, dict):
        node = {
            check_recursively_json(k, transformation): check_recursively_json(v, transformation)
            for k, v in node.items()
        }

    return transformation(node)


def recompose_notebook(source: str, notebook_path: str, storage=None) -> NotebookNode:
    try:
        if not source.endswith('\n'):
            source += '\n'
        result_json_with_stubs, config = NotebookVisitor().visit(notebook_grammar.parse(source))
    except ParseError as e:
        raise HTTPError(500, f'Could not parse notebook:\n {e}')
    if storage is None:
        storage = identify_storage_from_config(config, notebook_path=notebook_path)
    result_json = replace_stub_nodes_with_real_outputs(result_json_with_stubs, storage)
    nb = nbformat.from_dict(result_json)
    nbformat.validate(nb)
    return nb


def replace_stub_nodes_with_real_outputs(json, storage):
    # traverse notebook, keep hashes
    hashes = []

    def store_id(node):
        if isinstance(node, NewpyterOutputPlaceholder):
            hashes.extend(node.hashes)
        return node

    result_json = check_recursively_json(json, transformation=store_id)
    # download content for hashes
    hash2json_output = storage.get_decrypted_fragments(hashes)

    # traverse once again, replace hashes with normal outputs
    def replace_with_output(node):
        if isinstance(node, NewpyterOutputPlaceholder):
            result = []
            for hash in node.hashes:
                result.extend(hash2json_output[hash])
            return result
        return node

    result_json = check_recursively_json(result_json, transformation=replace_with_output)
    return result_json


def identify_storage_from_config(config: List[Tuple[str, str]], notebook_path: str):
    config_dict = {}
    for k, v in config:
        config_dict.setdefault(k, []).append(v)

    try:
        version = ast.literal_eval(config_dict.pop('format', ["'v1'"])[0])
        assert version == 'v1', f'version: {version} not recognized'

        storages = config_dict.pop('storage', [])
        if len(config_dict) != 0:
            raise RuntimeError(f'Unknown config params: {config_dict}')

        if len(storages) == 0:
            return get_storage_for_notebook(notebook_filename=Path(notebook_path))
        if len(storages) > 1:
            raise NotImplementedError('Multiple storages not supported yet')

        return get_storage_for_notebook(
            notebook_filename=Path(notebook_path),
            storage_url=ast.literal_eval(storages[0])
        )
    except:
        print(f'Failed to identify storage, \n found config: {config}')
        raise


def collapse_list_of_strings(list_or_str):
    if isinstance(list_or_str, str):
        return list_or_str
    assert isinstance(list_or_str, list)
    return '\n'.join(list_or_str)


def decompose_notebook(notebook: NotebookNode, storage: AbstractStorage) -> str:
    output_lines = []
    output_lines += ["# newpyter.format: 'v1'"]
    if storage.url is not None:
        output_lines += [f'# newpyter.storage: {storage.url!r}']
    with storage.get_upload_context() as storage_context:
        for cell in notebook['cells']:
            output_lines.extend(decompose_cell(cell, storage_context))

    assert all(isinstance(x, str) for x in output_lines)
    return '\n'.join(output_lines) + '\n'


def normalize_cell_outputs(input: List[Dict]) -> List[Dict]:
    if not isinstance(input, List):
        return input
    if any(not isinstance(d, Dict) for d in input):
        return input
    return [
        dict(sorted(d.items())) for d in input
    ]


def decompose_cell(cell, storage_context: StorageContext) -> List[str]:
    cell_type = cell['cell_type']
    cell_output_lines = []
    if cell_type == "markdown":
        cell_output_lines.append(encode_cell_text(collapse_list_of_strings(cell['source']), cell_type='md'))
    elif cell_type == 'raw':
        cell_output_lines.append(encode_cell_text(collapse_list_of_strings(cell['source']), cell_type='raw'))
    else:
        assert cell_type == 'code'
        count = cell.get('execution_count', '')
        assert count is None or isinstance(count, (str, int)), (count, type(count))
        if count is None:
            count = ''
        assert count == '' or isinstance(count, int)

        cell_output_lines.append(f'# In[{count}]')
        if len(cell.get('outputs', [])) > 0:
            outputs = cell['outputs']
            if len(outputs) == 1 and 'newpyter_not_found_hash' in outputs[0].get('metadata', {}):
                # special case when hash was not downloaded, so a special insertion was made
                output_hash = outputs[0]['metadata']['newpyter_not_found_hash']
            else:
                outputs = normalize_cell_outputs(cell['outputs'])
                output_hash = storage_context.store_fragment(outputs)
            cell_output_lines.append(f"# Out: {output_hash}")
        else:
            cell_output_lines.append(f'# Out: empty')

        if isinstance(cell['source'], str):
            cell_output_lines.append(cell['source'])
        else:
            assert isinstance(cell['source'], list)
            cell_output_lines.extend(cell['source'])
        # add an empty line in the end to visually separate from the previous cell
        cell_output_lines.append('')
    return cell_output_lines



__author__ = 'Alex Rogozhnikov'

from setuptools import setup

setup(
    name="newpyter",
    version='0.5.2',
    packages=['newpyter', 'newpyter.storage', 'newpyter.newpyviewer'],
    install_requires=[
        # useful for grammars
        'parsimonious',
        'nbformat',
        'sh',
        # download / upload to aws
        'boto3',
        # to parse configuration
        'toml',
        # for exception types, but it should be installed by jupyter
        'tornado',
    ],
    extras_require={
        'viewer': ["fastapi~=0.85.1", "nbconvert~=7.2.2"]
    },
    entry_points={
        'console_scripts': [
            'newpyter=newpyter.__main__:main',
        ]
    },
)


import json
import logging
import typing
from contextlib import contextmanager
from hashlib import sha1
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Dict, Set, Tuple, Any
from warnings import warn
from urllib.parse import urlparse

import boto3
import sh
from tornado.web import HTTPError

from newpyter.utils import remove_prefix, remove_suffix

logger = logging.getLogger(__file__)

FragmentType = Any

def get_hash(content: bytes) -> str:
    """ Default way is to use SHA1 hashing """
    return sha1(content).hexdigest()


# keeps track of everything uploaded during this session.
# Stores pairs (url, hash)
# This storage is persistent during single app start
uploaded_path_hashes: Set[Tuple[str, str]] = set()


class Encryptor:
    """ Simple symmetric encryption """

    def __init__(self, password=None):
        assert password is None, 'Encryption is to be implemented yet'

    def encrypt(self, x: bytes) -> bytes:
        return x

    def decrypt(self, x: bytes) -> bytes:
        return x


class StorageContext:
    """
    Accumulates elements in cache
    so that later they could be sent to cloud/storage in batch
    """

    def __init__(self, encryptor: Encryptor):
        self.cache: Dict[str, bytes] = {}
        self.encryptor = encryptor

    def store_fragment(self, fragment: FragmentType) -> str:
        assert isinstance(fragment, list)
        fragment = json.dumps(fragment).encode('utf-8')
        assert isinstance(fragment, bytes)
        fragment = self.encryptor.encrypt(fragment)
        fragment_hash = get_hash(fragment)
        self.cache[fragment_hash] = fragment
        return fragment_hash


def output_for_missing_fragment(hash):
    return [{
        "execution_count": None,
        "output_type": "execute_result",
        "data": {"text/plain": f"Couldn't find output for {hash}"},
        "metadata": {"newpyter_not_found_hash": hash},
    }]


class AbstractStorage:
    def __init__(self, encryptor: Encryptor, local_cache: Path):
        self.url: str # should be assigned in implementations
        self.encryptor = encryptor
        self.local_cache = local_cache
        self.notebook_specific_caches = local_cache.joinpath('temp_folders')
        self.notebook_specific_caches.mkdir(exist_ok=True, parents=True)
        self.uploaded_hashes: Set[str] = set()

    @contextmanager
    def get_upload_context(self):
        upload_context_manager = StorageContext(self.encryptor)
        yield upload_context_manager
        self.store_fragments(upload_context_manager.cache)

    def store_fragments(self, fragments: Dict[str, bytes]):
        # save locally
        for hash, encrypted_fragment in fragments.items():
            if not self.local_cache.joinpath(hash).exists():
                with self.local_cache.joinpath(hash).open('wb') as f:
                    f.write(encrypted_fragment)

        # collect which fragments were not uploaded yet
        for_upload = {
            hash: encrypted_fragment
            for hash, encrypted_fragment in fragments.items()
            if (self.url, hash) not in uploaded_path_hashes
        }
        if len(for_upload) > 0:
            try:
                self._upload_fragments(for_upload)
                uploaded_path_hashes.update([(self.url, hash) for hash in for_upload])
            except BaseException as e:
                raise HTTPError(500, message=f'Failed to upload jupyter outputs: \n{e}')

    def _upload_fragments(self, for_upload: Dict[str, bytes]):
        raise NotImplementedError("Storing is not implemented")

    def _download_fragments(self, hashes: List[str]) -> typing.Iterable[bytes]:
        raise NotImplementedError("Loading is not implemented in storage")

    def get_decrypted_fragments(self, hashes: List[str]) -> Dict[str, FragmentType]:
        hash2encrypted_fragment = {}
        hashes_for_downloading = []
        for hash in set(hashes):
            if self.local_cache.joinpath(hash).exists():
                with self.local_cache.joinpath(hash).open('rb') as f:
                    hash2encrypted_fragment[hash] = f.read()
            else:
                hashes_for_downloading.append(hash)

        if len(hashes_for_downloading) > 0:
            try:
                encrypted_fragments = self._download_fragments(hashes=hashes_for_downloading)
                hash2encrypted_fragment.update(dict(zip(hashes_for_downloading, encrypted_fragments)))
            except BaseException as e:
                logger.error(e)

        result = {
            hash: json.loads(self.encryptor.decrypt(encrypted_fragment).decode('utf-8'))
            for hash, encrypted_fragment
            in hash2encrypted_fragment.items()
        }

        for hash in hashes:
            if hash not in result:
                # would be better to return error cell, but it does not allow adding metadata
                result[hash] = output_for_missing_fragment(hash)
        return result


class MemoryMockStorage(AbstractStorage):
    url = ':MemoryMockStorage:'

    def __init__(self):
        """ Storage used for testing purposed only """
        self.stored: Dict[str, bytes] = {}
        self.encryptor = Encryptor()
        self.return_empty = False
        # does not init parent, as local cache is not used

    def store_fragments(self, fragments: Dict[str, bytes]):
        return self.stored.update(fragments)

    def get_decrypted_fragments(self, hashes: List[str]) -> Dict[str, FragmentType]:
        if self.return_empty:
            return {hash: output_for_missing_fragment(hash) for hash in hashes}
        return {
            hash: json.loads(self.encryptor.decrypt(self.stored[hash]).decode('utf-8'))
            for hash in hashes
        }

    def __repr__(self):
        return "<MemoryMockStorage>"


class SSHStorage(AbstractStorage):
    def __init__(self, ssh_path: str, encryptor: Encryptor, local_cache: Path, ssh_key: None):
        super().__init__(encryptor=encryptor, local_cache=local_cache)

        possible_prefixes = ['ssh://', 'scp://']
        assert any(ssh_path.startswith(prefix) for prefix in possible_prefixes)
        for prefix in possible_prefixes:
            ssh_path = remove_prefix(ssh_path, prefix)

        ssh_path = remove_suffix(ssh_path, '/') + '/'
        self.url = ssh_path
        self.ssh_key = ssh_key
        self.key_params: list = []
        if self.ssh_key is not None:
            self.key_params = ['-i', Path(self.ssh_key).expanduser()]

    def _upload_fragments(self, for_upload: Dict[str, bytes]):
        with TemporaryDirectory(dir=self.notebook_specific_caches) as temp_dir:
            for hash, fragment in for_upload.items():
                with open(Path(temp_dir).joinpath(hash), 'wb') as f:
                    f.write(fragment)
                # TODO use something better than scp
                sh.scp(*self.key_params, '-r', Path(temp_dir).joinpath(hash), self.url + hash)

    def _download_fragments(self, hashes: List[str]) -> typing.Iterator[bytes]:
        for hash in hashes:
            path_on_this_machine = self.local_cache.joinpath(hash)
            if not path_on_this_machine.exists():
                # TODO use something better than scp
                sh.scp(*self.key_params, self.url + hash, path_on_this_machine)
                # this hash is already there and will not be uploaded
                uploaded_path_hashes.add((self.url, hash))

            with path_on_this_machine.open('rb') as f:
                yield f.read()

    def __repr__(self):
        return f"<SSHStorage at {self.url}, local={self.local_cache}>"

class LocalStorage(AbstractStorage):
    def __init__(self, encryptor: Encryptor, local_cache: Path):
        warn("Local Storage is meant only for testing, your outputs won't be available to anyone but you!")
        super().__init__(encryptor=encryptor, local_cache=local_cache)

    def _upload_fragments(self, for_upload: Dict[str, bytes]):
        pass

    def _download_fragments(self, hashes: List[str]) -> List[bytes]:
        # should look for if not found in local cache
        raise RuntimeError(f'Fragments were not found: {hashes}')

    def __repr__(self):
        return f"<LocalStorage at {self.local_cache}>"



class S3Storage(AbstractStorage):
    def __init__(self, s3_path: str, encryptor: Encryptor, local_cache: Path, **boto3_parameters):
        super().__init__(encryptor=encryptor, local_cache=local_cache)

        possible_prefixes = ['s3://']
        self.boto3_parameters = boto3_parameters
        assert any(s3_path.startswith(prefix) for prefix in possible_prefixes)
        self.url = remove_suffix(s3_path, '/') + '/'
        parsing_result = urlparse(self.url, allow_fragments=False)
        self.bucket, path_in_bucket = parsing_result.netloc, parsing_result.path
        assert path_in_bucket.startswith('/')
        self.path_in_bucket = path_in_bucket[1:]

    def _upload_fragments(self, for_upload):
        # dump cache to local directory, then use aws s3 sync
        client = boto3.client('s3', **self.boto3_parameters)
        logger.info(f'Upload {list(for_upload)} to {self.bucket}{self.path_in_bucket}')
        with TemporaryDirectory(dir=self.notebook_specific_caches) as temp_dir:
            for hash, fragment in for_upload.items():
                local_path = Path(temp_dir).joinpath(hash)
                with open(local_path, 'wb') as f:
                    f.write(fragment)

                client.upload_file(str(local_path), self.bucket, self.path_in_bucket + hash)

    def _download_fragments(self, hashes: List[str]) -> List[bytes]:
        result = []
        client = boto3.client('s3', **self.boto3_parameters)
        logger.info(f'Download {hashes} from {self.bucket}{self.path_in_bucket}')
        for hash in hashes:
            path_on_this_machine = self.local_cache.joinpath(hash)
            if not path_on_this_machine.exists():
                try:
                    client.download_file(self.bucket, self.path_in_bucket + hash, str(path_on_this_machine))
                except BaseException as _:
                    logger.info(f'did not locate {self.path_in_bucket + hash} in bucket {self.bucket}, prefixing with / now')
                    # in old protocol versions, file was saved with a leading slash (by mistake),
                    # this isn't so since ~nov 2021, and after ~ apr 2023 this condition should be deleted
                    client.download_file(self.bucket, '/' + self.path_in_bucket + hash, str(path_on_this_machine))
                # this hash is already there and will not be uploaded
                uploaded_path_hashes.add((self.url, hash))
            with path_on_this_machine.open('rb') as f:
                result.append(f.read())
        return result

    def __repr__(self):
        return f"<S3Storage at {self.url}, local={self.local_cache}>"
